/**
 * Copyright 1998-2007, CHISEL Group, University of Victoria, Victoria, BC, Canada.
 * All rights reserved.
 */
package ca.uvic.csr.shrimp.SearchBean;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.Serializable;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;

import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableModel;


/**
 * This search strategy provides a threaded facility similar to
 * searching tools embedded in text editors and/or file finders.
 *
 * @author Jingwei Wu, Jeff Michaud
 */
public class GrepSearchStrategy extends SearchStrategy implements Serializable {
	
	public static final String NAME = "Source Code Search";
	private JComboBox patBox;
	private JCheckBox caseChk;

	private JButton searchBtn;

	private JTable table;
	private JScrollPane scroller;

	private JButton browseBtn;
	private JLabel selectLabel;
	private JLabel resultLabel;

	private JPanel inputPanel;
	private JPanel outputPanel;

	private Vector searchFiles;
	private Vector searchResults;
	private URI searchDir;
	private URI presentDir;

	// whether the search is case sensitive.
	private boolean caseSensitive;

	// The thread deals with searching.
	private SearchThread searching = null;

	private static final int COL_FILENAME = 0;
	private static final int COL_LINE_NUMBER = 1;
	private static final int COL_LINE_STRING = 2;

	/**
	 * Constructs a simple grep search strategy.
	 * @param searchDir 
	 * @param presentDir 
	 */
	public GrepSearchStrategy(URI searchDir, URI presentDir) {
		super(NAME);
		createComponentUI();
		this.searchDir = searchDir;
		this.presentDir = presentDir;
	}

	public void setDirectories(URI searchDir, URI presentDir) {
		this.searchDir = searchDir;
		this.presentDir = presentDir;
	}
	

	
	public void setSearchText(String text) {
		if (patBox != null) {
			patBox.setSelectedItem(text);
		}
	}

	/**
	 * Creates the component UI supporting this SearchStrategy.
	 * @return the component UI for this SearchStrategy.
	 */
	protected JComponent createComponentUI() {
		JPanel panel = new JPanel();
		panel.setLayout(new BorderLayout());
		panel.add(getInputPanel(), BorderLayout.NORTH);
		panel.add(getOutputPanel(), BorderLayout.CENTER);
		compUI = panel;
		return panel;
	}

	/**
	 * Creates the input panel.
	 */
	private JPanel getInputPanel() {
		if (inputPanel == null) {
			GridBagLayout gbl;
			GridBagConstraints gbc;
			gbl = new GridBagLayout();
			gbc = new GridBagConstraints();
			gbc.insets = new Insets(5, 5, 5, 5);

			inputPanel = new JPanel();
			inputPanel.setLayout(gbl);

			/////////////////////////////////////
			patBox = new JComboBox();
			patBox.setEditable(true);
			patBox.setMaximumRowCount(5);

			gbc.anchor = GridBagConstraints.WEST;
			gbc.fill = GridBagConstraints.HORIZONTAL;
			gbc.gridwidth = 1;
			inputPanel.add(new JLabel("Search pattern: "), gbc);
			inputPanel.add(patBox, gbc);
			gbc.gridwidth = GridBagConstraints.REMAINDER;
			inputPanel.add(getSearchButton(), gbc);

			/////////////////////////////////////
			gbc.anchor = GridBagConstraints.WEST;
			inputPanel.add(getCaseCheckBox(), gbc);
			gbc.fill = GridBagConstraints.HORIZONTAL;
			gbc.gridwidth = GridBagConstraints.REMAINDER;

			/*
			gbc.anchor = GridBagConstraints.CENTER;
			gbc.fill = GridBagConstraints.HORIZONTAL;
			gbc.gridwidth = GridBagConstraints.REMAINDER;
			inputPanel.add(new JSeparator(), gbc);
			*/
		}
		return inputPanel;
	}

	/**
	 * Creates the output panel.
	 */
	private JPanel getOutputPanel() {
		if (outputPanel == null) {
			outputPanel = new JPanel();
			outputPanel.setLayout(new BorderLayout());

			JPanel north;
			north = new JPanel();
			north.setLayout(new GridLayout(1, 2));

			JPanel panel;
			panel = new JPanel();
			panel.setLayout(new FlowLayout(FlowLayout.LEFT));
			panel.add(getBrowseButton());
			panel.add(new JLabel("Item:"));
			selectLabel = new JLabel("");
			panel.add(selectLabel);
			north.add(panel);

			panel = new JPanel();
			panel.setLayout(new FlowLayout(FlowLayout.RIGHT));
			resultLabel = new JLabel("");
			panel.add(resultLabel);
			north.add(panel);

			scroller = new JScrollPane(new JList());

			outputPanel.add(north, BorderLayout.NORTH);
			outputPanel.add(scroller, BorderLayout.CENTER);
		}
		return outputPanel;
	}

	/**
	 * Performs the required search.
	 */
	private void doSearch() {
		String input;

		// Do nothing if search text is empty.
		input = (String) patBox.getSelectedItem();
		if (input == null)
			return;

		patBox.addItem(input);

		if (searching != null)
			searching.tryStop();
		searchFiles = new Vector();
		getFiles(searchDir);
		searching = new SearchThread(input, searchFiles);
		searching.start();
	}

	/**
	 * Displays search results.
	 */
	private void displaySearchResults() {
		String text;
		if (searchResults == null || searchResults.size() == 0) {
			text = "Found nothing!";
			selectLabel.setText("");
			browseBtn.setEnabled(false);
		} else if (searchResults.size() == 1) {
			text = "Found 1 item!";
		} else {
			text = "Found " + searchResults.size() + " items!";
		}
		resultLabel.setText(text);

		Object obj;
		String name;
		Vector names = new Vector();
		for (int i = 0; i < searchResults.size(); i++) {
			obj = searchResults.elementAt(i);
			try {
				name = getSelectedDataReflector().getName(obj);
				names.add(name);
			} catch (IllegalArgumentException illegal) {
				illegal.printStackTrace();
			}
		}

		TableModel dataModel = new AbstractTableModel() {
			public int getRowCount() {
				return searchResults.size();
			}
			public int getColumnCount() {
				return 3;
			}
			public String getColumnName(int col) {
				if (col == COL_FILENAME)
					return "File";
				if (col == COL_LINE_NUMBER)
					return "Line Number";
				if (col == COL_LINE_STRING)
					return "Line";
				return null;
			}
			public Object getValueAt(int row, int col) {
				Vector result = (Vector) searchResults.elementAt(row);

				//File - set it to the relative path of the source directory
				if (col == COL_FILENAME) {
					String fpath = (String) result.elementAt(col);
					return fpath.substring(searchDir.toString().length());
				}

				//Line Number
				if (col == COL_LINE_NUMBER)
					return result.elementAt(col);

				//Line			
				if (col == COL_LINE_STRING) {
					String line = (String) result.elementAt(col);
					return line.trim();
				}

				return null;
			}
		};

		table = new JTable(dataModel);
		table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		ListSelectionModel rowSM = table.getSelectionModel();
		rowSM.addListSelectionListener(new ListSelectionListener() {
			public void valueChanged(ListSelectionEvent e) {
				browseBtn.setEnabled(true);
			}
		});

		table.addMouseListener(new MouseAdapter() {
			public void mouseClicked(MouseEvent e) {
				if (e.getClickCount() == 2) {
					getBrowseButton().doClick();
				}
				selectLabel.setText("" + (table.getSelectedRow() + 1));
			}
		});

		outputPanel.remove(scroller);
		scroller = new JScrollPane(table);
		outputPanel.add(scroller, BorderLayout.CENTER);
		outputPanel.validate();
		outputPanel.repaint();
	}

	/**
	 * Creates the case-sensitive checkbox.
	 */
	private JCheckBox getCaseCheckBox() {
		if (caseChk == null) {
			caseChk = new JCheckBox("Case Sensitive");
			caseChk.addItemListener(new ItemListener() {
				public void itemStateChanged(ItemEvent e) {
					int state = e.getStateChange();
					if (state == ItemEvent.SELECTED) {
						caseSensitive = true;
					} else {
						caseSensitive = false;
					}
				}
			});
		}
		return caseChk;
	}

	/**
	 * Creates the search button.
	 */
	protected JButton getSearchButton() {
		if (searchBtn == null) {
			searchBtn = new JButton("Search");
			searchBtn.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					doSearch();
				}
			});
		}
		return searchBtn;
	}

	/**
	 * Creates the browse button that cause a BrowseActionEvent
	 * be fired to listeners.
	 */
	private JButton getBrowseButton() {
		if (browseBtn == null) {
			browseBtn = new JButton("Browse");
			browseBtn.setEnabled(false);
			browseBtn.setMargin(new Insets(3, 5, 2, 4));
			browseBtn.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					//Build the URL from what we know		
					int index = table.getSelectedRow();
					String uriStr = (String) table.getValueAt(index, 0);
					String line = (String) table.getValueAt(index, 1);
					if (uriStr.lastIndexOf('/') <= 1)
					    uriStr = "/~default~" + uriStr;
					uriStr = uriStr.replace('.', '_');
					uriStr = presentDir.toString() + uriStr + ".html" + "#" + line;
					try {
						URI uri = new URI(uriStr);
						BrowseActionEvent bae = new BrowseActionEvent(GrepSearchStrategy.this, uri);
						fireBrowseActionEvent(bae);
					} catch (URISyntaxException ex) {
                        ex.printStackTrace();
                    }
				}
			});
		}
		return browseBtn;
	}

	/**
	 * Processes the given <code>data</code> that is going to
	 * be displayed on the compoent UI of this SearchStrategy.
	 * The new search results replaces the old search results.
	 * @param data the data to be examined. Each object in the
	 * collection is treated as raw data and must be checked by
	 * the data reflector in this SearchStrategy. If the data
	 * is <code>null</code> or does not contain any object, all
	 * the search results displayed in the component UI must be
	 * cleaned properly.
	 * @see #getSelectedDataReflector()
	 */
	public void processUpdate(Collection data) {
		searchResults.clear();
		processAppend(data);
	}

	/**
	 * Processes the given <code>data</code> that is going to
	 * be displayed on the compoent UI of this SearchStrategy.
	 * The new search results will be appended after the old
	 * search results.
	 * @param data the data to be examined. Each object in the
	 * collection is treated as raw data and must be checked by
	 * the data reflector in this SearchStrategy.
	 * @see #getSelectedDataReflector()
	 */
	public void processAppend(Collection data) {
		String input;
		input = (String) patBox.getSelectedItem();
		if (input == null)
			return; // Do nothing.

		if (searching != null)
			searching.tryStop();
		searching = new SearchThread(input, searchFiles);
		searching.start();
	}

	/**
	 * Returns whether or not this strategy has a search running
	 */
	public boolean isSearching() {
		return (searching != null && !searching.isDone());
	}

	/**
	 * This class is created for threaded searching.
	 */
	class SearchThread extends Thread {
		private Vector files;
		private boolean isDone = false;
		private String pattern;

		public SearchThread(String pattern, Vector files) {
			this.files = files;
			this.pattern = pattern;
		}

		public void run() {
			isDone = false;

			// set the cursor to busy
			Container container = searchBtn.getTopLevelAncestor();
			Cursor cursor = container.getCursor();
			container.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

			//"grep" files for pattern
			searchResults = new Vector();
			for (Iterator iter = files.iterator(); iter.hasNext();) {
                String filename = (String) iter.next();
                boolean match = false;
                int lineNumber = 1;
                String matchingLine = "";
        		try {
                    BufferedReader reader = new BufferedReader (new FileReader (filename));               
                    String line = reader.readLine();
                    while (line != null && !match) {
                	    match = Matcher.match(line, pattern, Matcher.CONTAINS_MODE, caseSensitive);
                	    if (match) {
                	        matchingLine = line;
                	    }
                	    lineNumber++;
                	    line = reader.readLine();
                    }
                } catch (FileNotFoundException e) {
                    e.printStackTrace();
                } catch (IOException e) {
                    e.printStackTrace();
                }
                if (match) {
                    Vector result = new Vector (3);
                    result.add(COL_FILENAME, filename);
                    result.add(COL_LINE_NUMBER, "" + lineNumber);
                    result.add(COL_LINE_STRING, matchingLine);
                    searchResults.add(result);
                }
            }
			
			//Display results
			displaySearchResults();

			// inform listeners that search is complete
			fireSearchCompletedEvent();
			// set the cursor back
			container.setCursor(cursor);

			isDone = true;
		}

		public boolean isDone() {
			return isDone;
		}

		public void tryStop() {
			searching = null;
		}
	} //inner class SearchThread

	/**
	* Retrieves the java files to be searched
	*/
	private void getFiles(URI dir) {
	    //TODO add URI's to searchFiles list
	    //searchFiles.add(file.getPath());
	}

} 
