/**
 * Copyright 1998-2007, CHISEL Group, University of Victoria, Victoria, BC, Canada.
 * All rights reserved.
 */
 
package ca.uvic.csr.shrimp.jambalaya.ProtegeDataBean;

import java.awt.Component;
import java.awt.GridLayout;
import java.util.Enumeration;
import java.util.Vector;

import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;

import ca.uvic.csr.shrimp.PanelModeConstants;
import ca.uvic.csr.shrimp.DataBean.AbstractArtifact;
import ca.uvic.csr.shrimp.DataBean.Artifact;
import ca.uvic.csr.shrimp.DataBean.DataBean;
import edu.stanford.smi.protege.model.Cls;
import edu.stanford.smi.protege.model.Facet;
import edu.stanford.smi.protege.model.Frame;
import edu.stanford.smi.protege.model.Instance;
import edu.stanford.smi.protege.model.KnowledgeBase;
import edu.stanford.smi.protege.model.Project;
import edu.stanford.smi.protege.model.Slot;
import edu.stanford.smi.protege.util.ComponentUtilities;
import edu.stanford.smi.protege.util.DoubleClickListener;
import edu.stanford.smi.protege.widget.AbstractSlotWidget;
import edu.stanford.smi.protege.widget.ClsWidget;
import edu.stanford.smi.protege.widget.FormWidget;


/**
 * The ProtegeArtifact is a container for a single entity of 
 * information in the Shrimp system.  It represents a class
 * or instance in a Protege Knowledgebase.
 *
 * @author Casey Best & Rob Lintern
 */

public class ProtegeArtifact extends AbstractArtifact {

	/** These constants are customized panels that this artifact knows how to handle */
	public final static String PANEL_FORM = "Form";
	public final static String PANEL_METRICS = "Metrics";
	//public final static String PANEL_SEARCH_CRITERIA = "Search Panel";

	/** These constants are for customized data metrics */
    protected final static String INHERITORS_KEY = "Metrics - Inheritors";
	protected final static String INSTANCES_KEY = "Metrics - Instances";
	protected final static String DEPTH_KEY = "Metrics - Tree Depth";
	protected final static String STRAHLER_KEY = "Metrics - Strahler";

	// The frame this artifact represents
	protected Frame protegeFrame;

	protected ProtegeArtifact(DataBean db, String name, String type, Object externalId){
		super (db, name, type, externalId);
	}
    
	public ProtegeArtifact(DataBean db, Frame frame) {
		this (db, ProtegeDataBean.frameToArtifactName(frame), ((ProtegeDataBean)db).frameToArtifactType(frame), ProtegeDataBean.frameToExternalArtifactID(frame));
        this.protegeFrame = frame;  
	}
    
	/**
	 * Returns the Protege frame this artifact represents.
	 */
	public Frame getFrame() {
		return protegeFrame;
	}
	
	/**
	 * This method has been overridden to ask the frame for specific slot info if
	 * there isn't already data for the given attribute name.  
	 * NOTE: If there is data in the attributes hash table, the slot WON'T be checked.
	 * @param attrName The name of the attribute or slot you want the value for.
	 */
	public Object getAttribute(String attrName) {
		Object attrValue = super.getAttribute(attrName);
		if ((attrName != null) && (attrValue == null) && (protegeFrame != null)) {
			Project project = ((ProtegeDataBean)dataBean).getProject();
			if (project != null) {
				KnowledgeBase knowledgeBase = project.getKnowledgeBase();
				Slot slot = knowledgeBase.getSlot(attrName);
				if (slot != null) {
					attrValue = protegeFrame.getOwnSlotValue(slot);
				}
			}
		}
		return attrValue;
	}

	
	/* (non-Javadoc)
	 * @see ca.uvic.csr.shrimp.DataBean.Artifact#getDefaultPanelModeOrder()
	 */
	public String[] getDefaultPanelModeOrder() {
		String[] ret = new String[2];
		ret[0] = PanelModeConstants.CHILDREN;
		ret[1] = PanelModeConstants.PANEL_DEFAULT;
		
		return ret;
	}
	
	/**
	 * Creates the panel for the panel mode passed in.  Returns null if there isn't
	 * a panel for the mode.
	 *
	 * @param mode The panel mode to create a panel for.
	 */
	protected Component createPanel (String mode) {
		if (mode.equals (PanelModeConstants.PANEL_DEFAULT)) {
			boolean addWidgetPanel = true;
				
			if (addWidgetPanel) {
				addWidgetPanel();
			}
		} else if (mode.equals (PANEL_FORM)) {
			addWidgetPanel();
		} else if (mode.equals (PanelModeConstants.PANEL_ANNOTATION)) {
			addAnnotationPanel();
		} else if (mode.equals (PANEL_METRICS)) {
			addMetricsPanel();
		} else if (mode.equals (PanelModeConstants.PANEL_ATTRIBUTES)) {
		    addAttributesPanel();
		}
		
		JComponent panel = (JComponent)customizedPanels.get(mode);
		return panel;
	}
	
	/**
	 * Returns the available customized panels for this artifact.
	 */
	public Vector getCustomizedPanelNames (){
		Vector names = new Vector();
		names.addElement (PanelModeConstants.PANEL_ANNOTATION);
 		if (protegeFrame != null) {
			names.addElement (PANEL_FORM);
			names.addElement (PANEL_METRICS);
			names.addElement (PanelModeConstants.PANEL_ATTRIBUTES);
			// for caBIO
			/*
			if (protegeFrame instanceof SimpleInstance) {
				Cls directType = ((SimpleInstance)protegeFrame).getDirectType();
				if (directType != null && directType.getName().toLowerCase().indexOf("searchcriteria") != -1) { 
					names.addElement (PANEL_SEARCH_CRITERIA);
				}
			}
			*/
		}

		return names;
	}


	
	/********************* Methods from Object ***********************/
	
	/**
	 * Overrides the Object method with the same name.  Used to index
	 * hashTables correctly.
	 */
	public int hashCode() {
		/* ***** NOTE: We dont want to use id.hashCode here because id is a string, and
		 * if the String's length exceeds 16 characters, String.hashCode() samples only a 
		 * portion of the String, which may result in duplicate hashcodes.
		 */
		int hashCode = (protegeFrame == null) ? 0 : protegeFrame.getFrameID().hashCode();
		return hashCode;
	}
	
	/** 
	 * @see ca.uvic.csr.shrimp.DataBean.AbstractArtifact#equals(java.lang.Object)
	 */
	public boolean equals(Object obj) {
		// TODO should take into account hashcode returned above
		return super.equals(obj);
	}

	
	/**
	 * Returns a clone of this artifact
	 */
	public Object clone() {
		ProtegeArtifact art = new ProtegeArtifact (dataBean, protegeFrame);
		Enumeration keys = getCustomizedPanelNames().elements();
		while (keys.hasMoreElements()) {
			String name = (String)keys.nextElement();
			art.setCustomizedPanel (name, getCustomizedPanel (name));
		}

		keys = getAttributeNames().elements();
		while (keys.hasMoreElements()) {
			String name = (String)keys.nextElement();
			art.setAttribute (name, getAttribute (name));
		}
		
		return art;
	}
	
	/*********** The Panels available in this domain ****************/
	/*
	protected JScrollPane addSearchPanel() {
		// get the a search criteria instance associated with this object's type
		// in future will be slot, for now just hard code
		JPanel pnl = new JPanel ();
		pnl.setLayout(new BorderLayout ());
		KnowledgeBase kb = protegeFrame.getKnowledgeBase();
		//final String searchCriteriaInstanceName = protegeFrame.getName();
		final Instance searchCriteriaInstance = (Instance) protegeFrame; //searchCriteriaCls.createDirectInstance(null);
		Cls searchCriteriaCls = searchCriteriaInstance.getDirectType();
		final String searchCriteriaClassName = searchCriteriaCls.getName();
		final String searcherObjectName  = (String) searchCriteriaCls.getOwnSlotValue(kb.getSlot("Preferred_Name"));
		final ClsWidget widget = getWidgetPanel (searchCriteriaInstance); 
		if (widget != null) {
			pnl.add((Component)widget, BorderLayout.CENTER);
			// take out any non search criteria slot widgets
			try {
				Collection descendentComps = ComponentUtilities.getDescendentsOfClass(Class.forName ("edu.stanford.smi.protege.widget.AbstractSlotWidget"), (Component)widget);
				SearchCriteria searchCriteriaJavaClass = (SearchCriteria) Class.forName("gov.nih.nci.caBIO.bean." + searchCriteriaClassName).newInstance();
				for (Iterator iter = descendentComps.iterator(); iter.hasNext();) {
					AbstractSlotWidget abstractSlotWidget = (AbstractSlotWidget) iter.next();
					abstractSlotWidget.setBorder(BorderFactory.createLineBorder(Color.black));
					final Slot slot = abstractSlotWidget.getSlot();
					JButton btnShowValues = new JButton ("+");
					btnShowValues.addActionListener(new ActionListener () {
						public void actionPerformed(ActionEvent e) {
							Set values = getPotentialSlotValues (slot);	
							ChoosePotentialValueDialog choosePotentialValueDialog = new ChoosePotentialValueDialog (values);
							if (choosePotentialValueDialog.isAccepted()) {
								Object value = choosePotentialValueDialog.getChosenValue();
								searchCriteriaInstance.setOwnSlotValue(slot, value);
							}
						}
					});
					abstractSlotWidget.add(btnShowValues, BorderLayout.EAST);
					
					if (!isValidSearchCriteriaSlot(searchCriteriaInstance, searchCriteriaJavaClass, slot)) {
						abstractSlotWidget.setVisible(false);
					}
				}
			} catch (Exception e1) {
				e1.printStackTrace();
			}
			
		}
		JButton btnSearch = new JButton ("Search");
		pnl.add(btnSearch, BorderLayout.SOUTH);
		btnSearch.addActionListener(new ActionListener () {
			public void actionPerformed(ActionEvent e) {
				search (widget, searcherObjectName, searchCriteriaClassName, searchCriteriaInstance);
			}
		});
		JScrollPane scrollPane = new JScrollPane(pnl);
		//setCustomizedPanel (PANEL_SEARCH_CRITERIA, scrollPane);
		return scrollPane;
	}
	
	private void search (ClsWidget widget, String searcherObjectName, String searchCriteriaClassName, Instance searchCriteriaInstance) {
		KnowledgeBase kb = protegeFrame.getKnowledgeBase();
		try {
			SearchCriteria searchCriteriaJavaClass = (SearchCriteria) Class.forName("gov.nih.nci.caBIO.bean." + searchCriteriaClassName).newInstance();
			// now match slots to properties
			Collection ownSlots = searchCriteriaInstance.getOwnSlots();
			Set searchCriteriaProperties = BeanPropertyDialog.getValidProperties(true, true, true, searchCriteriaJavaClass);
			Map searchCriteriaPropNameToProp = getPropNameToPropMap(searchCriteriaProperties);
			for (Iterator iter = ownSlots.iterator(); iter.hasNext();) {
				Slot slot = (Slot) iter.next();
				if (!isValidSearchCriteriaSlot(searchCriteriaInstance, searchCriteriaJavaClass, slot)) continue;
				String slotName = slot.getName();
				Object slotValue = searchCriteriaInstance.getOwnSlotValue(slot);
				
				PropertyDescriptor property = (PropertyDescriptor)searchCriteriaPropNameToProp.get(slotName);
				if (property != null && slotValue != null) {
					try {
						Method setter = property.getWriteMethod();
						if (slotValue instanceof String || slotValue instanceof Number) {
							if (slotValue instanceof Number) {
								slotValue = new Long (slotValue.toString());
							}
							System.out.println("\nsetting slotName: " + slotName);
							System.out.println("slotValue: " + slotValue);
							// use the setter of this property 
							Object [] args = {slotValue};
							setter.invoke(searchCriteriaJavaClass, args);
						} 
					} catch (Exception e) {
						e.printStackTrace();
					}
				}
			}					
			try {
				Class searcherClass = Class.forName("gov.nih.nci.caBIO.bean." + searcherObjectName);
				Object searcherObject = searcherClass.newInstance();
				Class [] classArgs = {SearchCriteria.class};
				Method searchMethod = searcherClass.getMethod("search", classArgs); // find the "search (SearchCriteria)" nethod for this object
				Object [] objArgs = {searchCriteriaJavaClass};
				SearchResult searchResult = (SearchResult) searchMethod.invoke(searcherObject, objArgs);
				System.out.println("searchResult: " + searchResult);
				Cls searchCriteriaCls = searchCriteriaInstance.getDirectType();
				Slot hasResultSlot = kb.getSlot("HAS_ASSOCIATED_RESULT");
				// find the associated result class
				Cls resultCls = null;
				if (hasResultSlot != null) {
					resultCls = (Cls) searchCriteriaCls.getOwnSlotValue(hasResultSlot);
				} 
				if (searchResult != null) {
					Object [] resultSet = searchResult.getResultSet();
					for (int i = 0; i < resultSet.length; i++) {
						Object resultObj = resultSet[i];
						//System.out.println("resultObj[" + i + "]: " + resultObj);
						// now create instance that matches this result in the knowledge base
						if (resultCls != null) {
							// create some children of this class, filling in the appropriate fields
							Instance resultInstance = resultCls.createDirectInstance(null);
							// match the properties of the result object with the slots of this new instance
							Set searchResultProperties = BeanPropertyDialog.getValidProperties(true, false, false, resultObj);
							for (Iterator iter = searchResultProperties.iterator(); iter.hasNext();) {
								PropertyDescriptor searchResultProperty = (PropertyDescriptor) iter.next();
								String resultPropertyName = searchResultProperty.getDisplayName();
								System.out.println("\nresultPropertyName: " + resultPropertyName);
								// see if there is slot that corresponds to this property
								Slot resultSlot = kb.getSlot(resultPropertyName);
								if (resultSlot != null && resultInstance.getOwnSlots().contains(resultSlot)) {
									//get the value of this property from our result object
									Method getter = searchResultProperty.getReadMethod();
									Object [] args = {};
									Object returnValue = getter.invoke(resultObj, args);
									//set the value of the slot
									if (returnValue != null && ((returnValue instanceof String) || (returnValue instanceof Long))) {
										if (returnValue instanceof Long) {
											returnValue = new Integer (returnValue.toString());
										}
										resultInstance.setOwnSlotValue(resultSlot, returnValue);
										System.out.println("valid slot and valid value: " + returnValue);
									} else {
										System.out.println("valid slot but not valid value");
									}
								} else {
									System.out.println("not a valid slot");
								}
							}
						}
					}
				}
			} catch (Exception e) {
				e.printStackTrace();
			} 
		} catch (Exception e1) {
				e1.printStackTrace();
		}
			
	}

	private boolean isValidSearchCriteriaSlot (Instance searchCriteriaInstance, SearchCriteria searchCriteriaJavaClass, Slot slot) {
		boolean b = true;
		Set properties = BeanPropertyDialog.getValidProperties(true, true, true, searchCriteriaJavaClass);
		Map propNameToProp = getPropNameToPropMap(properties);
		PropertyDescriptor property = (PropertyDescriptor)propNameToProp.get(slot.getName());		
		b = property != null;
		
		return b;
	}
	
	private Map getPropNameToPropMap (Set properties) {
		Map propNameToProp = new HashMap();
		for (Iterator iter = properties.iterator(); iter.hasNext();) {
			PropertyDescriptor property = (PropertyDescriptor) iter.next();
			propNameToProp.put(property.getDisplayName(), property);
		}
		return propNameToProp;
	}
	

	private class ChoosePotentialValueDialog extends JDialog {
		JComboBox cmbxValues;
		JButton btnOk = new JButton ("OK");
		JButton btnCancel = new JButton ("Cancel");
		boolean accepted = false;
		Object chosenValue = null;
		
		public ChoosePotentialValueDialog (Set values) {
			super (ApplicationAccessor.getApplication().getParentFrame(), "Choose Value", true);
			getContentPane().setLayout(new BorderLayout ());
			cmbxValues = new JComboBox (new Vector (values));
			cmbxValues.setEditable(false);
			getContentPane().add(cmbxValues, BorderLayout.CENTER);
			btnOk.addActionListener(new ActionListener () {
				public void actionPerformed(ActionEvent e) {
					accepted = true;
					chosenValue = cmbxValues.getSelectedItem();
					dispose();
				}
			});
			btnCancel.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					accepted = false;
					chosenValue = null;
					dispose();
				}
			});
			JPanel pnl = new JPanel ();
			pnl.add(btnOk);
			pnl.add(btnCancel);
			getContentPane().add(pnl, BorderLayout.SOUTH);
			pack();	
			Toolkit toolKit = Toolkit.getDefaultToolkit();
			Dimension screenSize = toolKit.getScreenSize();
			int x = screenSize.width/2 - ChoosePotentialValueDialog.this.getWidth()/2;
			int y = screenSize.height/2 - ChoosePotentialValueDialog.this.getHeight()/2;
			setLocation(x,y);
			setVisible(true);			
		}
		
		public boolean isAccepted () {
			return accepted;
		}
		
		public Object getChosenValue () {
			return chosenValue;
		}
	}

	private Set getPotentialSlotValues (Slot slot) {
		Set values = new HashSet();
		// query caBIO for potential values
		values.add("a");
		values.add("b");
		values.add("c");
		return values;
	}
	*/
	
	private ClsWidget getWidgetPanel (Frame frame) {
		ClsWidget widget = null;
		if (frame != null && (frame instanceof Instance) && (dataBean instanceof ProtegeDataBean)) {
			widget = ((ProtegeDataBean) dataBean).getClsWidget (frame);
			
			if (widget instanceof FormWidget) {
				FormWidget formWidget = (FormWidget)widget;
				try {
					// add double click listeners to all the appropriate widgets
					Vector descendentComponents = new Vector (ComponentUtilities.getDescendentsOfClass(Class.forName ("edu.stanford.smi.protege.widget.AbstractSlotWidget"), formWidget));
					for (int i = 0; i < descendentComponents.size(); i++) {
						AbstractSlotWidget abstractSlotWidget = (AbstractSlotWidget) descendentComponents.elementAt(i);
						abstractSlotWidget.setDoubleClickListener (new DoubleClickAdapter (dataBean, this));
					}
				} catch (ClassNotFoundException e) {
					System.out.println ("Couldn't find the class: edu.stanford.smi.protege.widget.AbstractSlotWidget");
					e.printStackTrace();
				}
			}
		}
		return widget;
	}
	
	/**
	 * Add the widget pane
	 */
	protected void addWidgetPanel () {
		ClsWidget widget = getWidgetPanel (protegeFrame); 
		if (widget != null) {
			JScrollPane scrollPane = new JScrollPane((Component)widget);
			setCustomizedPanel (PanelModeConstants.PANEL_DEFAULT, scrollPane);
			setCustomizedPanel (PANEL_FORM, scrollPane);
		}
	}
	
	/**
	 * This adapter handles the double click of an item in any widget in protege
	 */
	private class DoubleClickAdapter implements DoubleClickListener {
		private ProtegeDataBean dataBean;
		private AbstractArtifact sourceArt;
		
		public DoubleClickAdapter (DataBean dataBean, Artifact sourceArt) {
			this.dataBean = (ProtegeDataBean)dataBean;
			this.sourceArt = (AbstractArtifact)sourceArt;
		}
		
		public void onDoubleClick (Object item) {
			if (!(item instanceof Slot) && !(item instanceof Facet) /*&& !(item instanceof Cls && ((Cls)item).isMetaCls() )*/) {
				Artifact target =  dataBean.findArtifactByExternalId(ProtegeDataBean.frameToExternalArtifactID((Frame)item));
				if (target != null) {
					fireCustomizedPanelEvent (new ProtegeDoubleClickEvent (target, sourceArt));
				} else {
					fireCustomizedPanelEvent (new ProtegeDoubleClickEvent (null, sourceArt));
				}
			}
		}
	}
	
	
	protected void addMetricsPanel () {
		JPanel pane = new JPanel ();
		
	    String inheritorsString = "# Inheritors: ";
	    String instancesString = "# Instances: ";
	    String depthString = "Tree Depth: ";
	    String strahlerString = "Strahler: ";
	    
		ProtegeMetrics metrics = ((ProtegeDataBean)dataBean).getMetrics();
        setAttribute(INHERITORS_KEY, new Integer (metrics.getNumInheritors(protegeFrame)));
        setAttribute(INSTANCES_KEY, new Integer (metrics.getNumInstances(protegeFrame)));
        setAttribute(DEPTH_KEY, new Integer (metrics.getTreeDepth(protegeFrame)));
        setAttribute(STRAHLER_KEY, new Integer (metrics.getStrahler(protegeFrame)));
		
		strahlerString += getAttribute(STRAHLER_KEY);
       	inheritorsString += getAttribute(INHERITORS_KEY);
        instancesString += getAttribute(INSTANCES_KEY);
        depthString += getAttribute(DEPTH_KEY);
		
	    JLabel inheritorsLabel;
	    JLabel instancesLabel;
	    JLabel depthLabel;
	    JLabel strahlerLabel;
        inheritorsLabel = new JLabel(inheritorsString);
        instancesLabel = new JLabel(instancesString);
        depthLabel = new JLabel(depthString);
        strahlerLabel = new JLabel(strahlerString);
        pane.setLayout(new GridLayout(0, 1));
        pane.add(inheritorsLabel);
        pane.add(instancesLabel);
        pane.add(depthLabel);
        pane.add(strahlerLabel);
		setCustomizedPanel (PANEL_METRICS, pane);
	}

    public int getTreeDepth() {
        int depth = 0;
        ProtegeMetrics pm = ((ProtegeDataBean)dataBean).getMetrics();
        if (protegeFrame instanceof Cls) {
            Cls cls = (Cls) protegeFrame;
            depth = pm.getTreeDepth(cls);
        } 
        return depth;
    }
	

}





